import asyncio
import base64
import subprocess
import sys
import time
from pathlib import Path
from typing import AsyncIterator

import numpy as np
import pytest
from graphql import SourceLocation
from strawberry import Schema

from coniql import __version__
from coniql.app import create_schema

TEST_DIR = Path(__file__).resolve().parent

EXPECTED_SIM_SINE = [0.0, 2.938926261462366, 4.755282581475768]


@pytest.fixture(scope="session")
def schema():
    return create_schema(False)


@pytest.mark.asyncio
async def test_get_sim_sine(schema: Schema):
    query = """
query {
    getChannel(id: "ssim://sine(-5,5,10,1,60)") {
        value {
            float
            string
            base64Array {
                numberType
            }
            stringArray
        }
        display {
            widget
        }
    }
}
"""
    for i, x in enumerate(EXPECTED_SIM_SINE):
        if i != 0:
            await asyncio.sleep(1.0)
        result = await schema.execute(query)
        assert result.data == {
            "getChannel": {
                "value": {
                    "float": x,
                    "string": "%.5f" % x,
                    "base64Array": None,
                    "stringArray": None,
                },
                "display": {"widget": "TEXTUPDATE"},
            },
        }


@pytest.mark.asyncio
async def test_put_sim_sine_fails(schema: Schema):
    query = """
mutation {
    putChannels(ids: ["ssim://sine"], values: ["32"]) {
        value {
            float
        }
    }
}
"""
    result = await schema.execute(query)
    assert result.data is None
    assert result.errors is not None
    assert (
        result.errors[0].message
        == "Cannot put ['32'] to ['sine'], as they aren't writeable"
    )
    assert result.errors[0].locations == [SourceLocation(column=5, line=3)]
    assert result.errors[0].path == ["putChannels"]


@pytest.mark.asyncio
async def test_subscribe_sim_sine(schema: Schema):
    query = """
subscription {
    subscribeChannel(id: "ssim://sine") {
        value {
            float
        }
    }
}
"""
    results = []
    start = time.time()
    resp = await schema.subscribe(query)
    assert isinstance(resp, AsyncIterator)
    async for result in resp:
        results.append(result.data)
        if time.time() - start > 2:
            break
    for i, x in enumerate(EXPECTED_SIM_SINE):
        assert results[i] == {"subscribeChannel": {"value": {"float": x}}}
    assert len(results) == 3


@pytest.mark.asyncio
async def test_subscribe_ramp_wave(schema: Schema):
    query = """
subscription {
    subscribeChannel(id: "ssim://rampwave(3, 0.2)") {
        value {
            stringArray
        }
    }
}
"""
    results = []
    start = time.time()
    resp = await schema.subscribe(query)
    assert isinstance(resp, AsyncIterator)
    async for result in resp:
        results.append(result.data)
        if len(results) == 4:
            break
    # First result immediate, then takes 3x 0.2s
    assert time.time() - start - 0.6 < 0.2
    expected = [
        ["0.00000", "1.00000", "2.00000"],
        ["1.00000", "2.00000", "3.00000"],
        ["2.00000", "3.00000", "4.00000"],
        ["3.00000", "4.00000", "5.00000"],
    ]
    for i, x in enumerate(expected):
        assert results[i] == {"subscribeChannel": {"value": {"stringArray": x}}}


@pytest.mark.asyncio
async def test_get_sim_sinewave(schema: Schema):
    query = """
query {
    getChannel(id: "ssim://sinewave") {
        value {
            string
            stringArray(length: 10)
            base64Array(length: 100) {
                numberType
                base64
            }
        }
    }
}
"""
    result = await schema.execute(query)
    assert result.data == {
        "getChannel": {
            "value": {
                "string": str(np.zeros(50)),
                "stringArray": ["%.5f" % x for x in np.zeros(10)],
                "base64Array": {
                    "numberType": "FLOAT64",
                    "base64": base64.b64encode(np.zeros(50).tobytes()).decode(),
                },
            }
        }
    }
    await asyncio.sleep(1.0)
    result = await schema.execute(query)
    assert result.data is not None
    b64s = result.data["getChannel"]["value"]["base64Array"]["base64"]
    str_value = np.frombuffer(base64.b64decode(b64s), dtype=np.float64)
    value = base64.b64decode(b64s)
    assert result.data == {
        "getChannel": {
            "value": {
                "string": str(str_value),
                "stringArray": ["%.5f" % x for x in str_value[:10]],
                "base64Array": {
                    "numberType": "FLOAT64",
                    "base64": base64.b64encode(value).decode(),
                },
            }
        }
    }


def test_cli_version():
    cmd = [sys.executable, "-m", "coniql", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_schema_contains_no_snake_case(schema: Schema):
    """Test that the Schema contains no snake_case members which may need to be
    converted to camelCase (as per GraphQL convention).

    Currently we set auto_camel_case=False when constructing the Schema, as currently
    there are no variables that need to be converted.

    If this test fails due to additions to the schema, this can probably be deleted
    (as long as we also re-enable auto_camel_case)"""

    introspected_schema = schema.introspect()

    for type in introspected_schema["__schema"]["types"]:
        name = type["name"]
        # Some inbuilt types with leading doule underscores exist, e.g. __Schema, __Type
        if name[0:2] != "__":
            assert "_" not in name

        if type["fields"] is not None:
            for field in type["fields"]:
                assert "_" not in field["name"]



import asyncio
import json
import random
import string
import subprocess
import sys
from datetime import datetime, timedelta
from pathlib import Path
from typing import cast
from unittest.mock import ANY

import pytest
from aioca import purge_channel_caches
from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL
from strawberry.subscriptions.protocols.graphql_transport_ws.types import (
    ConnectionAckMessage,
    ConnectionInitMessage,
    SubscribeMessage,
    SubscribeMessagePayload,
)
from strawberry.subscriptions.protocols.graphql_ws import (
    GQL_CONNECTION_ACK,
    GQL_CONNECTION_INIT,
    GQL_START,
)
from strawberry.subscriptions.protocols.graphql_ws.types import (
    OperationMessage,
    StartPayload,
)

from coniql.app import create_app
from coniql.caplugin import CAPlugin, CASubscriptionManager
from coniql.strawberry_schema import store_global

SOFT_RECORDS = str(Path(__file__).parent / "soft_records.db")

PV_PREFIX = "".join(random.choice(string.ascii_uppercase) for _ in range(12)) + ":"

BASE64_0_1688_2 = {
    "numberType": "FLOAT64",
    "base64": "AAAAAAAAAAA1XrpJDAL7PwAAAAAAAABA",
}

SUBSCRIPTION_TIMEOUT = 10


def wait_for_ioc(ioc):
    while True:
        line = ioc.stdout.readline()
        if "complete" in line:
            return


@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    yield loop
    loop.close()


def ioc_creator(pv_prefix=PV_PREFIX) -> subprocess.Popen:
    process = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "epicscorelibs.ioc",
            "-m",
            f"P={pv_prefix}",
            "-d",
            SOFT_RECORDS,
        ],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )
    wait_for_ioc(process)
    return process


def ioc_cleanup(process):
    purge_channel_caches()
    try:
        process.communicate("exit()")
    except ValueError:
        # Someone else already called communicate
        pass


@pytest.fixture(scope="module")
def ioc():
    process = ioc_creator()
    yield process
    ioc_cleanup(process)


def check_put_timestamp(result):
    thens = [
        datetime.fromisoformat(r["time"]["datetime"]) for r in result["putChannels"]
    ]

    now = datetime.now()
    for then in thens:
        diff = now - then
        # Shouldn't take more than this time to get the result of a put out
        assert diff.total_seconds() < 0.2


@pytest.fixture(scope="function")
async def client(aiohttp_client):
    cors = True
    debug = False
    graphiql = False
    connection_init_wait_timeout = timedelta(seconds=2)
    client = await aiohttp_client(
        create_app(cors, debug, graphiql, connection_init_wait_timeout)
    )
    return client


longout_get_query = (
    """
query {
    getChannel(id: "ca://%slongout") {
        id
        value {
            float
            string
        }
        display {
            widget
            controlRange {
                min
                max
            }
            displayRange {
                min
                max
            }
            alarmRange {
                min
                max
            }
            warningRange {
                min
                max
            }
            units
            precision
            form
        }
        status {
            quality
        }
    }
}
"""
    % PV_PREFIX
)


longout_get_query_result = {
    "getChannel": {
        "id": f"ca://{PV_PREFIX}longout",
        "value": {"float": 42.0, "string": "42"},
        "display": {
            "widget": "TEXTINPUT",
            "controlRange": {"min": 10.0, "max": 90.0},
            "displayRange": {"min": 0.0, "max": 100.0},
            "alarmRange": {"min": 2.0, "max": 98.0},
            "warningRange": {"min": 5.0, "max": 96.0},
            "units": "",
            "precision": None,
            "form": None,
        },
        "status": {"quality": "VALID"},
    },
}


longout_str_get_query = (
    """
query {
    getChannel(id: "ca://%slongout.RTYP") {
        value {
            string
        }
    }
}
"""
    % PV_PREFIX
)


longout_str_get_query_result = {"getChannel": {"value": {"string": "longout"}}}

enum_get_query = (
    """
query {
    getChannel(id: "ca://%senum") {
        value {
            string
            float
        }
        display {
            choices
        }
    }
}
"""
    % PV_PREFIX
)


enum_get_query_result = {
    "getChannel": {
        "value": {"string": "nm", "float": 3.0},
        "display": {"choices": ["m", "mm", "um", "nm"]},
    }
}

nan_get_query = (
    """
query {
    getChannel(id: "ca://%snan") {
        value {
            float
        }
    }
}
"""
    % PV_PREFIX
)

nan_get_query_result = {"getChannel": {"value": {"float": None}}}

long_and_enum_put_query = """
mutation {
    putChannels(ids: ["ca://%slongout", "ca://%senum"], values: ["55", "1"]) {
        value {
            string
        }
        time {
            datetime
        }
    }
}
""" % (
    PV_PREFIX,
    PV_PREFIX,
)

long_and_enum_put_query_result = {
    "putChannels": [
        {"value": {"string": "55"}, "time": ANY},
        {"value": {"string": "mm"}, "time": ANY},
    ]
}

list_put_query = (
    """
mutation {
    putChannels(ids: ["ca://%swaveform"], values: ["[0, 1.688, 2]"]) {
        value {
            stringArray
            base64Array {
                numberType
                base64
            }
        }
        time {
            datetime
        }
    }
}
"""
    % PV_PREFIX
)

list_put_query_result = {
    "putChannels": [
        {
            "value": {
                "stringArray": ["0.0", "1.7", "2.0"],
                "base64Array": BASE64_0_1688_2,
            },
            "time": ANY,
        }
    ]
}

base64_put_query = """
mutation {
    putChannels(ids: ["ca://%swaveform"], values: [%s]) {
        value {
            stringArray
        }
        time {
            datetime
        }
    }
}
""" % (
    PV_PREFIX,
    json.dumps(json.dumps(BASE64_0_1688_2)),
)

base64_put_query_result = {
    "putChannels": [
        {
            "value": {"stringArray": ["0.0", "1.7", "2.0"]},
            "time": ANY,
        }
    ]
}


def get_longout_subscription_query(pv_prefix):
    return (
        """
subscription {
    subscribeChannel(id: "ca://%slongout") {
        value {
            float
        }
        status {
            quality
        }
    }
}
"""
        % pv_prefix
    )


longout_subscription_result = [
    {"subscribeChannel": {"value": {"float": 42.0}, "status": {"quality": "VALID"}}},
    {"subscribeChannel": {"value": None, "status": {"quality": "INVALID"}}},
]

ticking_subscription_query = (
    """
subscription {
    subscribeChannel(id: "ca://%sticking") {
        value {
            string(units: true)
            float
        }
        display {
            precision
            units
        }
    }
}
"""
    % PV_PREFIX
)


def get_ticking_subscription_result(startVal):
    return [
        {
            "subscribeChannel": {
                "value": {"string": f"{startVal}0000 mm", "float": startVal},
                "display": {"precision": 5, "units": "mm"},
            }
        },
        {
            "subscribeChannel": {
                "value": {"string": f"{startVal + 1}0000 mm", "float": startVal + 1},
                "display": None,
            }
        },
        {
            "subscribeChannel": {
                "value": {"string": f"{startVal + 2}0000 mm", "float": startVal + 2},
                "display": None,
            }
        },
    ]


subscribe_params = [
    (
        GRAPHQL_TRANSPORT_WS_PROTOCOL,
        ConnectionInitMessage().as_dict(),
        ConnectionAckMessage().as_dict(),
        SubscribeMessage(
            id="sub1",
            payload=SubscribeMessagePayload(query=ticking_subscription_query),
        ).as_dict(),
    ),
    (
        GRAPHQL_WS_PROTOCOL,
        OperationMessage(type=GQL_CONNECTION_INIT),
        OperationMessage(type=GQL_CONNECTION_ACK),
        OperationMessage(
            type=GQL_START,
            id="sub1",
            payload=StartPayload(query=ticking_subscription_query),
        ),
    ),
]


@pytest.fixture(
    scope="session",
    params=subscribe_params,
    ids=["graphql_transport_ws_protocol", "graphql_ws_protocol"],
)
def subscription_data(request):
    """Fixture for the possible subscription types"""
    return request.param


@pytest.fixture(autouse=True)
def clear_subscription_manager() -> None:
    """Reset the CASubscriptionManager inside the CAPlugin

    This ensures there's no record of PVs between tests"""

    ca_plugin: CAPlugin = cast(CAPlugin, store_global.plugins["ca"])
    ca_plugin.subscription_manager = CASubscriptionManager()



# This will eventually go in a different module so pvi and coniql (and gda) can use it
from enum import Enum

import strawberry


@strawberry.enum
class DisplayForm(str, Enum):
    """
    Instructions for how a number should be formatted for display
    """

    # Use the default representation from value
    DEFAULT = "DEFAULT"
    # Force string representation, most useful for array of bytes
    STRING = "STRING"
    # Binary, precision determines number of binary digits
    BINARY = "BINARY"
    # Decimal, precision determines number of digits after decimal point
    DECIMAL = "DECIMAL"
    # Hexadecimal, precision determines number of hex digits
    HEX = "HEX"
    # Exponential, precision determines number of digits after decimal point
    EXPONENTIAL = "EXPONENTIAL"
    # Exponential where exponent is multiple of 3, precision determines number
    # of digits after decimal point
    ENGINEERING = "ENGINEERING"


@strawberry.enum
class Widget(str, Enum):
    """
    Widget that should be used to display a Channel
    """

    # Editable text input
    TEXTINPUT = "TEXTINPUT"
    # Read-only text display
    TEXTUPDATE = "TEXTUPDATE"
    # Multiline read-only text display
    MULTILINETEXTUPDATE = "MULTILINETEXTUPDATE"
    # Read-only LED indicator
    LED = "LED"
    # Editable combo-box style menu for selecting between fixed choices
    COMBO = "COMBO"
    # Editable check box
    CHECKBOX = "CHECKBOX"
    # Editable progress type bar
    BAR = "BAR"
    # Clickable button to send default value to Channel
    BUTTON = "BUTTON"
    # X-axis for lines on a graph. Only valid within a Group with widget Plot
    PLOTX = "PLOTX"
    # Y-axis for a line on a graph. Only valid within a Group with widget Plot
    PLOTY = "PLOTY"


class Layout(str, Enum):
    SCREEN = "SCREEN"
    BOX = "BOX"
    PLOT = "PLOT"
    ROW = "ROW"
    TABLE = "TABLE"



# Support type hints for asyncio.Queue
from __future__ import annotations

import asyncio
import logging
import uuid
from asyncio import Event, Queue
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import AsyncIterator, Callable, Dict, List, Optional, Sequence

from aioca import (
    DBE_PROPERTY,
    FORMAT_CTRL,
    FORMAT_TIME,
    CANothing,
    Subscription,
    caget,
    cainfo,
    camonitor,
    caput,
)
from aioca.types import AugmentedValue

from coniql.coniql_schema import Widget
from coniql.metrics import update_subscription_metrics
from coniql.plugin import Plugin, PutValue
from coniql.types import (
    Channel,
    ChannelDisplay,
    ChannelFormatter,
    ChannelQuality,
    ChannelRole,
    ChannelStatus,
    ChannelTime,
    ChannelValue,
    Range,
)

coniql_logger = logging.getLogger(__name__)
TRANSPORT = "ca://"


class CAChannelMaker:
    def __init__(self, name, writeable: bool):
        self.name = name
        self.cached_status: Optional[ChannelStatus] = None
        self.formatter = ChannelFormatter()
        # No camonitor is capable of updating whether a channel is writeable,
        # so this value is immutable.
        self.writeable = writeable

    @staticmethod
    def _create_formatter(value: AugmentedValue) -> ChannelFormatter:
        formatter = ChannelFormatter()
        precision = getattr(value, "precision", 0)
        units = getattr(value, "units", "")
        if hasattr(value, "dtype"):
            # numpy array
            formatter = ChannelFormatter.for_ndarray(
                precision,
                units,
            )
        elif hasattr(value, "enums"):
            # enum
            formatter = ChannelFormatter.for_enum(value.enums)
        elif isinstance(value, (int, float)):
            # number
            formatter = ChannelFormatter.for_number(
                precision,
                units,
            )

        return formatter

    def channel_from_update(
        self,
        time_value: Optional[AugmentedValue] = None,
        meta_value: Optional[AugmentedValue] = None,
        send_quality: bool = False,
    ) -> Channel:
        """Create a Channel object, populated with data taken from the provided
        AugmentedValue(s).

        Specify `send_quality` to force the "quality" attribute to be included in
        the returned channel - by default it will only send this attribute if it
        has changed state.
        """
        value = None
        time = None
        status = None
        display = None

        if meta_value is not None and meta_value.ok:
            self.formatter = CAChannelMaker._create_formatter(meta_value)
            # The value itself should not have changed for a meta_value update,
            # but the formatter may have, so send an updated value.
            value = ChannelValue(meta_value, self.formatter)
            display = ChannelDisplay(
                description=self.name,
                role=ChannelRole.RW,
                widget=Widget.TEXTINPUT,
                form=None,
            )
            if hasattr(meta_value, "enums"):
                display.choices = meta_value.enums
            if hasattr(meta_value, "precision"):
                display.precision = meta_value.precision
            if hasattr(meta_value, "units"):
                display.units = meta_value.units
                display.controlRange = Range(
                    min=meta_value.lower_ctrl_limit,
                    max=meta_value.upper_ctrl_limit,
                )
                display.displayRange = Range(
                    min=meta_value.lower_disp_limit,
                    max=meta_value.upper_disp_limit,
                )
                display.alarmRange = Range(
                    min=meta_value.lower_alarm_limit,
                    max=meta_value.upper_alarm_limit,
                )
                display.warningRange = Range(
                    min=meta_value.lower_warning_limit,
                    max=meta_value.upper_warning_limit,
                )

        if time_value is not None:
            if time_value.ok:
                assert time_value.timestamp
                value = ChannelValue(time_value, self.formatter)
                quality = str(ChannelQuality(time_value.severity))
                if (
                    send_quality
                    or self.cached_status is None
                    or self.cached_status.quality != quality
                ):
                    status = ChannelStatus(
                        quality=quality,
                        message="",
                        mutable=self.writeable,
                    )
                    self.cached_status = status
                time = ChannelTime(
                    seconds=time_value.timestamp,
                    nanoseconds=time_value.raw_stamp[1],
                    userTag=0,
                )
            else:
                # An update where .ok is false indicates a disconnection.
                status = ChannelStatus(
                    quality="INVALID",
                    message="",
                    mutable=self.writeable,
                )
                self.cached_status = status
        id = TRANSPORT + self.name
        return CAChannel(id, value, time, status, display)


@dataclass
class CAChannel(Channel):
    id: Optional[str]
    value: Optional[ChannelValue]
    time: Optional[ChannelTime]
    status: Optional[ChannelStatus]
    display: Optional[ChannelDisplay]

    def get_id(self) -> Optional[str]:
        return self.id

    def get_time(self) -> Optional[ChannelTime]:
        return self.time

    def get_status(self) -> Optional[ChannelStatus]:
        return self.status

    def get_value(self) -> Optional[ChannelValue]:
        return self.value

    def get_display(self) -> Optional[ChannelDisplay]:
        return self.display


@dataclass
class CallbackContext:
    callback: Callable[[Channel], None]


@dataclass
class SubscriptionData:
    pv: str

    time_value: Optional[AugmentedValue]
    time_monitor: Subscription

    meta_value: Optional[AugmentedValue]
    meta_monitor: Subscription

    all_values_received: Event

    callbacks: Dict[str, CallbackContext]
    maker: CAChannelMaker

    subscribers: int


class DataEnum(Enum):
    TIME_VALUE = "time_value"
    META_VALUE = "meta_value"


class CASubscriptionManager:
    """Pools camonitor requests across all subscriptions, ensuring we only have one
    active subscription for each PV."""

    def __init__(self) -> None:
        self.pvs: Dict[str, SubscriptionData] = {}
        self.metrics_task: Optional[asyncio.Task] = None
        self.locks: Dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)

    async def update_metrics(self) -> None:
        value_monitor_last_dropped: Dict[str, int] = defaultdict(int)
        meta_monitor_last_dropped: Dict[str, int] = defaultdict(int)

        while True:
            for x in self.pvs.values():
                value_monitor_last_dropped[x.pv] = update_subscription_metrics(
                    x.time_monitor, value_monitor_last_dropped[x.pv], {"type": "value"}
                )
                meta_monitor_last_dropped[x.pv] = update_subscription_metrics(
                    x.meta_monitor, meta_monitor_last_dropped[x.pv], {"type": "meta"}
                )
            await asyncio.sleep(10)  # Metrics only need updating infrequently

    def __callback(self, pv: str, key: DataEnum, v: AugmentedValue):
        if pv not in self.pvs:
            return

        data = self.pvs[pv]

        if key == DataEnum.TIME_VALUE:
            data.time_value = v
        elif key == DataEnum.META_VALUE:
            data.meta_value = v
        else:
            raise KeyError(f"Unrecognised key {key}")

        # Ensure we have both values before creating the channel
        # This ensures the first update sent to the client has both
        # values and metadata
        if data.time_value is None or data.meta_value is None:
            return

        if not data.all_values_received.is_set():
            # The first subscription return is handled as part of the
            # `subscribe` function, which is blocked waiting on this event
            data.all_values_received.set()
        else:
            # Otherwise, construct the appropriate channel and call all callbacks
            if key == DataEnum.TIME_VALUE:
                channel = data.maker.channel_from_update(time_value=data.time_value)
            elif key == DataEnum.META_VALUE:
                channel = data.maker.channel_from_update(meta_value=data.meta_value)

            for context in data.callbacks.values():
                context.callback(channel)

    async def subscribe(
        self, pv: str, callback: Callable[[Channel], None], callback_key: str
    ):
        """Subscribe to the given PV.

        This function will block until both a time and meta value update has been
        received from the PV. Once the data is received the provided callback will be
        called immediately, with a Channel object that contains both time and meta
        values.

        Caller must provide a key that will be associated with the callback. This same
        key must be passed to the `unsubscribe` function."""
        callback_context = CallbackContext(callback)

        # Restrict access to the shared dictionary - otherwise issues arise if two
        # clients attempt to subscribe to the same PV at the same time
        async with self.locks[pv]:
            # One-time async init across all subscriptions
            if self.metrics_task is None:
                self.metrics_task = asyncio.create_task(self.update_metrics())

            # Either this PV is new or we've previously closed the monitors.
            if (
                pv not in self.pvs
                or self.pvs[pv].meta_monitor.state == Subscription.CLOSED
            ):
                # A specific request required for whether the channel is writeable.
                # This will not be updated, so wait until a callback is received
                # before making the request when the channel is likely be connected.
                # NOTE: This MUST happen before the camonitors are created, otherwise we
                # run the risk of the initial monitor callbacks happening while we're
                # waiting for this cainfo call to return.
                writeable = True
                try:
                    info = await cainfo(pv)
                    writeable = info.write
                except CANothing:
                    # Unlikely, but allow subscriptions to continue.
                    pass

                # Initialize camonitors for this PV
                value_monitor = camonitor(
                    pv,
                    lambda v: self.__callback(pv, DataEnum.TIME_VALUE, v),
                    format=FORMAT_TIME,
                    notify_disconnect=True,
                )

                # Monitor PV only for property changes. For EPICS < 3.15 this monitor
                # will update once on connection but will not subsequently be triggered.
                # https://github.com/dls-controls/coniql/issues/22#issuecomment-863899258
                meta_monitor = camonitor(
                    pv,
                    lambda v: self.__callback(pv, DataEnum.META_VALUE, v),
                    events=DBE_PROPERTY,
                    format=FORMAT_CTRL,
                )

                maker = CAChannelMaker(pv, writeable)

                self.pvs[pv] = SubscriptionData(
                    pv=pv,
                    time_value=None,
                    time_monitor=value_monitor,
                    meta_value=None,
                    meta_monitor=meta_monitor,
                    all_values_received=Event(),
                    callbacks={callback_key: callback_context},
                    maker=maker,
                    subscribers=1,
                )

            else:
                self.pvs[pv].subscribers += 1
                self.pvs[pv].callbacks[callback_key] = callback_context

            # Construct and send a channel with both time and meta values
            data = self.pvs[pv]
            await data.all_values_received.wait()
            channel = data.maker.channel_from_update(
                time_value=data.time_value,
                meta_value=data.meta_value,
                send_quality=True,
            )
            callback_context.callback(channel)

    def unsubscribe(self, pv: str, callback_key: str) -> None:
        """Unsubscribe from the given PV. The callback key must be provided and must
        match the one passed to the `subscribe` function."""
        data = self.pvs[pv]

        data.subscribers -= 1

        data.callbacks.pop(callback_key)

        if data.subscribers == 0:
            data.time_monitor.close()
            data.meta_monitor.close()


class CAPlugin(Plugin):
    def __init__(self):
        self.subscription_manager = CASubscriptionManager()

    async def get_channel(self, pv: str, timeout: float) -> Channel:
        time_value, meta_value, info = await asyncio.gather(
            caget(pv, format=FORMAT_TIME, timeout=timeout),
            caget(pv, format=FORMAT_CTRL, timeout=timeout),
            cainfo(pv, timeout=timeout),
        )
        maker = CAChannelMaker(pv, info.write)
        return maker.channel_from_update(time_value=time_value, meta_value=meta_value)

    async def put_channels(
        self, pvs: List[str], values: Sequence[PutValue], timeout: float
    ):
        await caput(pvs, values, timeout=timeout)

    async def subscribe_channel(self, pv: str) -> AsyncIterator[Channel]:
        value: Queue[Channel] = asyncio.Queue(maxsize=1)

        # Generate unique key for this subscription
        uid = str(uuid.uuid4())

        def __callback(channel: Channel):
            # Ensure queue is empty
            while True:
                try:
                    value.get_nowait()
                except asyncio.QueueEmpty:
                    break

            value.put_nowait(channel)

        await self.subscription_manager.subscribe(pv, __callback, uid)

        try:
            while True:
                yield await value.get()

        finally:
            self.subscription_manager.unsubscribe(pv, uid)



import logging
from argparse import ArgumentParser
from datetime import timedelta
from typing import Any, Optional

import aiohttp_cors
from aiohttp import web
from aiohttp.hdrs import METH_GET, METH_POST
from strawberry.aiohttp.views import GraphQLView
from strawberry.schema.config import StrawberryConfig
from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL

import coniql.strawberry_schema as schema
from coniql.metrics import (
    MetricsExtension,
    MetricsGraphQLTransportWSHandler,
    MetricsGraphQLWSHandler,
    MetricsSchema,
    handle_metrics,
    metrics_middleware,
)

from . import __version__


def create_schema(debug: bool):
    # Disabling automatic camel casing saves ~1% CPU time as Strawberry won't attempt
    # to check/change our field names. Currently we don't have any snake_case fields
    # so there isn't any transformation happening anyway.
    config = StrawberryConfig(auto_camel_case=False)

    # Create the schema
    return MetricsSchema(
        query=schema.Query,
        subscription=schema.Subscription,
        mutation=schema.Mutation,
        extensions=[MetricsExtension],
        config=config,
    )


class GraphQLViewExtension(GraphQLView):
    """Use custom handlers to enable inprogress metrics for subscriptions"""

    graphql_transport_ws_handler_class = MetricsGraphQLTransportWSHandler
    graphql_ws_handler_class = MetricsGraphQLWSHandler


def create_app(
    use_cors: bool,
    debug: bool,
    graphiql: bool,
    connection_init_wait_timeout: Optional[timedelta] = None,
):
    # Create the schema
    strawberry_schema = create_schema(debug)

    kwargs: Any = {}
    if connection_init_wait_timeout:
        kwargs["connection_init_wait_timeout"] = connection_init_wait_timeout

    # Create the GraphQL view to attach to the app
    view = GraphQLViewExtension(
        schema=strawberry_schema,
        subscription_protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL],
        graphiql=graphiql,
        **kwargs,
    )

    # Create app
    app = web.Application(middlewares=[metrics_middleware])

    # Add routes
    app.router.add_route(METH_GET, "/ws", view)
    app.router.add_route(METH_POST, "/ws", view)
    app.router.add_route(METH_POST, "/graphql", view)

    app.router.add_route(METH_GET, "/metrics", handle_metrics)

    # Enable CORS for all origins on all routes (if applicable)
    if use_cors:
        cors = aiohttp_cors.setup(app)
        for route in app.router.routes():
            allow_all = {
                "*": aiohttp_cors.ResourceOptions(
                    allow_headers=("*"), max_age=3600, allow_credentials=True
                )
            }
            cors.add(route, allow_all)

    return app


def configure_logger(debug: bool = False, fmt: Optional[str] = None) -> None:
    class OptionalTraceFormatter(logging.Formatter):
        def __init__(self, debug: bool = False, fmt: Optional[str] = None) -> None:
            self.debug = debug
            super().__init__(fmt)

        def formatStack(self, stack_info: str) -> str:
            """Option to suppress the stack trace output"""
            if not self.debug:
                return ""
            return super().formatStack(stack_info)

    # Handler to print to stderr
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG if debug else logging.ERROR)
    console.setFormatter(OptionalTraceFormatter(debug, fmt))

    # Attach it to both Coniql and Strawberry loggers
    strawberry_logger = logging.getLogger("strawberry")
    strawberry_logger.addHandler(console)
    coniql_logger = logging.getLogger("coniql")
    coniql_logger.addHandler(console)


def main(args=None) -> None:
    """
    Entry point of the application.
    """
    parser = ArgumentParser(description="CONtrol system Interface over graphQL")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--cors",
        action="store_true",
        default=False,
        help="Allow CORS for all origins and routes",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Print stack trace on errors",
    )
    parser.add_argument(
        "--graphiql",
        action="store_true",
        default=False,
        help="Enable GraphiQL for testing at localhost:8080/ws",
    )
    parsed_args = parser.parse_args(args)

    logger_fmt = "[%(asctime)s::%(name)s::%(levelname)s]: %(message)s"
    configure_logger(parsed_args.debug, logger_fmt)

    app = create_app(parsed_args.cors, parsed_args.debug, parsed_args.graphiql)
    web.run_app(app)



import asyncio
import base64
import datetime
import json
from typing import AsyncGenerator, List, Optional, Sequence, Set, Union

import numpy as np
import strawberry

from coniql.caplugin import CAPlugin
from coniql.plugin import Plugin, PluginStore
from coniql.simplugin import SimPlugin
from coniql.types import Base64Array as TypeBase64Array
from coniql.types import Channel as TypeChannel
from coniql.types import ChannelDisplay
from coniql.types import ChannelStatus as TypeChannelStatus
from coniql.types import ChannelTime as TypeChannelTime
from coniql.types import ChannelValue as TypeChannelValue
from coniql.types import TypeFloatAlias

store_global = PluginStore()
store_global.add_plugin("ssim", SimPlugin())
store_global.add_plugin("ca", CAPlugin(), set_default=True)


def resolve_float(root: TypeChannelValue) -> Optional[float]:
    return root.formatter.to_float(root.value)


def resolve_string(root: TypeChannelValue, units: bool = False) -> Optional[str]:
    if units:
        return root.formatter.to_string_with_units(root.value)
    else:
        return root.formatter.to_string(root.value)


def resolve_base64Array(
    root: TypeChannelValue, length: int = 0
) -> Optional[TypeBase64Array]:
    return root.formatter.to_base64_array(root.value, length)


def resolve_stringArray(root: TypeChannelValue, length: int = 0) -> Optional[List[str]]:
    return root.formatter.to_string_array(root.value, length)


@strawberry.type
class ChannelValue:
    """
    Value that can be formatted in a number of ways
    """

    # The current value formatted as a Float, Null if not expressable
    float: Optional[TypeFloatAlias] = strawberry.field(resolver=resolve_float)
    # The current value formatted as a string
    string: Optional[str] = strawberry.field(resolver=resolve_string)
    # Array of base64 encoded numbers, Null if not expressable
    base64Array: Optional[TypeBase64Array] = strawberry.field(
        resolver=resolve_base64Array
    )
    # Array of strings, Null if not expressable
    stringArray: Optional[List[str]] = strawberry.field(resolver=resolve_stringArray)


@strawberry.type
class ChannelStatus(TypeChannelStatus):
    """
    The current status of a Channel, including alarm and connection status.
    Inherit directly from TypeChannelStatus.
    """

    pass


def resolver_datetime(root: TypeChannelTime) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(root.seconds)


@strawberry.type
class ChannelTime(TypeChannelTime):
    """
    Timestamp indicating when a value was last updated.
    Inherit fields from TypeChannelTime.
    """

    @strawberry.field
    async def datetime(root: TypeChannelTime) -> datetime.datetime:
        """The timestamp as a datetime object"""
        return datetime.datetime.fromtimestamp(root.seconds)


def resolver_id(root: "DeferredChannel") -> Optional[str]:
    channel = root.get_channel()
    return channel.get_id()


def resolver_value(root: "DeferredChannel") -> Optional[TypeChannelValue]:
    channel = root.get_channel()
    return channel.get_value()


def resolver_time(root: "DeferredChannel") -> Optional[TypeChannelTime]:
    channel = root.get_channel()
    return channel.get_time()


def resolver_status(root: "DeferredChannel") -> Optional[TypeChannelStatus]:
    channel = root.get_channel()
    return channel.get_status()


def resolver_display(root: "DeferredChannel") -> Optional[ChannelDisplay]:
    channel = root.get_channel()
    return channel.get_display()


@strawberry.type
class Channel(TypeChannel):
    """
    A single value with associated time, status and metadata. These values
    can be Null so that in a subscription they are only updated on change
    """

    def __init__(self, plugin_channel: TypeChannel):
        self.parentChannel = plugin_channel

    def get_id(self) -> Optional[str]:
        return self.parentChannel.get_id()

    def get_value(self) -> Optional[TypeChannelValue]:
        return self.parentChannel.get_value()

    def get_time(self) -> Optional[TypeChannelTime]:
        return self.parentChannel.get_time()

    def get_status(self) -> Optional[TypeChannelStatus]:
        return self.parentChannel.get_status()

    def get_display(self) -> Optional[ChannelDisplay]:
        return self.parentChannel.get_display()

    # ID that uniquely defines this Channel, normally a PV
    id: Optional[str] = strawberry.field(resolver=resolver_id)
    # The current value of this channel
    value: Optional[ChannelValue] = strawberry.field(
        resolver=resolver_value
    )  # type: ignore
    # When was the value last updated
    time: Optional[ChannelTime] = strawberry.field(
        resolver=resolver_time
    )  # type: ignore
    # Status of the connection, whether is is mutable, and alarm info
    status: Optional[ChannelStatus] = strawberry.field(
        resolver=resolver_status
    )  # type: ignore
    # How should the Channel be displayed
    display: Optional[ChannelDisplay] = strawberry.field(resolver=resolver_display)


class DeferredChannel(Channel):
    id: str
    lock: asyncio.Lock
    channel: Optional[Channel] = None

    async def populate_channel(self) -> Channel:
        raise NotImplementedError(self)

    def get_channel(self) -> Channel:
        assert self.channel
        return self.channel


class GetChannel(DeferredChannel):
    def __init__(self, channel_id: str, timeout: float, store: PluginStore):
        self.plugin, self.id = store.plugin_config_id(channel_id)
        # Remove the transport prefix from the read pv
        self.pv = store.transport_pv(channel_id)[1]
        self.timeout = timeout
        self.lock = asyncio.Lock()

    async def populate_channel(self) -> Channel:
        channel = await self.plugin.get_channel(self.pv, self.timeout)
        # Convert types.Channel object to a Strawberry schema Channel
        strawberry_channel = Channel(channel)
        self.channel = strawberry_channel
        return strawberry_channel


async def get_channel(id: strawberry.ID, timeout: float = 5.0) -> Channel:
    channel = GetChannel(id, timeout, store_global)
    await channel.populate_channel()
    return channel


@strawberry.type
class Query:
    # Get the current value of a Channel
    getChannel: Channel = strawberry.field(resolver=get_channel)


class SubscribeChannel(DeferredChannel):
    def __init__(self, channel_id: str, channel: Channel):
        self.id = channel_id
        self.channel = channel


async def subscribe_channel(id: strawberry.ID) -> AsyncGenerator[Channel, None]:
    """Subscribe to changes in top level fields of Channel,
    if they haven't changed they will be Null"""
    store: PluginStore = store_global
    plugin, channel_id = store.plugin_config_id(id)
    # Remove the transport prefix from the read pv
    pv = store.transport_pv(id)[1]
    async for channel in plugin.subscribe_channel(pv):
        # Convert types.Channel object to a Strawberry schema Channel
        strawberry_channel = Channel(channel)
        yield SubscribeChannel(channel_id, strawberry_channel)


@strawberry.type
class Subscription:
    """Tell mypy to ignore this line as it complains 'expression
    has type "T", variable has type "Channel"'. Possibly a Strawberry issue
    as strawberry.subscription is a partial function"""

    subscribeChannel: Optional[Channel] = strawberry.subscription(
        resolver=subscribe_channel
    )  # type: ignore


@strawberry.type
class Mutation:
    @strawberry.mutation
    async def putChannels(
        self,
        ids: List[strawberry.ID],
        values: List[str],
        timeout: float = 5.0,
    ) -> Sequence[Channel]:
        """Put a list of values to a list of Channels"""
        store: PluginStore = store_global
        pvs: List[str] = []
        plugins: Set[Plugin] = set()
        put_values = values
        for id in ids:
            plugin, channel_id = store.plugin_config_id(id)
            pv = channel_id
            assert pv, f"{channel_id} is configured read-only"
            plugins.add(plugin)
            pvs.append(store.transport_pv(pv)[1])
        results = []
        for value in put_values:
            put_value: Union[str, np.ndarray] = value
            if value[:1] in "[{":
                # need to json decode
                put_value = json.loads(value)
                if isinstance(put_value, dict):
                    # decode base64 array
                    dtype = np.dtype(put_value["numberType"].lower())
                    value_b = base64.b64decode(put_value["base64"])
                    # https://stackoverflow.com/a/6485943
                    put_value = np.frombuffer(value_b, dtype=dtype)
            results.append(put_value)
        assert len(results) == len(pvs), "Mismatch in ids and values length"
        assert (
            len(plugins) == 1
        ), "Can only put to pvs with the same transport, not %s" % [
            p.transport for p in plugins
        ]
        await plugins.pop().put_channels(pvs, results, timeout)
        channels: Sequence[GetChannel] = [
            GetChannel(channel_id, timeout, store) for channel_id in ids
        ]
        for channel in channels:
            await channel.populate_channel()
        return channels



import base64
import math
import time
from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Any, Callable, List, Optional

import numpy as np
import strawberry

from .coniql_schema import DisplayForm, Widget


@strawberry.type
@dataclass
class Range:
    """
    A range of numbers. Null in either field means unbounded in that direction.
    A value is in range if min <= value <= max
    """

    # The minimum number that is in this range
    min: float
    # The maximum that is in this range
    max: float

    def contains(self, value: float) -> bool:
        rmin = math.nan if self.min is None else self.min
        rmax = math.nan if self.max is None else self.max
        return rmin <= value <= rmax


class ChannelQuality(IntEnum):
    """
    Indication of how the current value of a Channel should be interpreted

    Note: The values align to the "menuAlarmSevr" enum in EPICS, with a couple of
    extension values.
    """

    # Value is known, valid, nothing is wrong
    VALID = 0
    # Value is known, valid, but is in the range generating a warning
    WARNING = 1
    # Value is known, valid, but is in the range generating an alarm condition
    ALARM = 2
    # Value is known, but not valid, e.g. a RW before its first put
    INVALID = 3
    # The value is unknown, for instance because the channel is disconnected
    UNDEFINED = 4
    # The Channel is currently in the process of being changed
    CHANGING = 5

    def __str__(self):
        # Returns the Enum item as a string e.g. "VALID", rather than integer value
        return self.name


# Map from display form to DisplayForm enum
DISPLAY_FORM_MAP = [
    DisplayForm.DEFAULT,
    DisplayForm.STRING,
    DisplayForm.BINARY,
    DisplayForm.DECIMAL,
    DisplayForm.HEX,
    DisplayForm.EXPONENTIAL,
    DisplayForm.ENGINEERING,
]

# GraphQL schema has a variable named 'float', which is an inbuilt type in Python
# and can lead to problems when used in type hinting. Create alias inbuilt float type.
TypeFloatAlias = float


@strawberry.type
class ChannelDisplay:
    # A human readable possibly multi-line description for a tooltip
    description: str
    # What access role does the Channel have
    role: "ChannelRole"
    # Default widget to display this Channel
    widget: Optional[Widget]
    # If numeric, the range the put value should be within
    controlRange: Optional[Range] = None
    # If numeric, the range the current value should be within
    displayRange: Optional[Range] = None
    # If numeric, the range outside of which an alarm will be produced
    alarmRange: Optional[Range] = None
    # If numeric, the range outside of which a warning will be produced
    warningRange: Optional[Range] = None
    # If numeric, the physical units for the value field
    units: Optional[str] = None
    # If numeric, the number of decimal places to display
    precision: Optional[int] = None
    # If numeric, how should value be displayed
    form: Optional[DisplayForm] = None
    # If given, the value should be one of these choices
    choices: Optional[List[str]] = None


@strawberry.enum
class NumberType(Enum):
    INT8 = "INT8"
    UINT8 = "UINT8"
    INT16 = "INT16"
    UINT16 = "UINT16"
    INT32 = "INT32"
    UINT32 = "UINT32"
    INT64 = "INT64"
    UINT64 = "UINT64"
    FLOAT32 = "FLOAT32"
    FLOAT64 = "FLOAT64"


@strawberry.enum
class ChannelRole(Enum):
    """
    What access role has the Channel
    """

    RO = "RO"
    WO = "WO"
    RW = "RW"


@strawberry.type
@dataclass
class Base64Array:
    # Type of the native array
    numberType: NumberType
    # Base64 encoded version of the array
    base64: str


def make_number_format_string(precision: Optional[int]) -> str:
    assert precision is not None
    return "{:.%df}" % precision


def return_if_number(value):
    if math.isfinite(value):
        return value


def return_none(*args, **kwargs):
    return None


class ChannelFormatter:
    @classmethod
    def for_number(
        cls, precision: Optional[int], units: Optional[str]
    ) -> "ChannelFormatter":
        number_format_string = make_number_format_string(precision)
        if units:
            units_format_string = f"{number_format_string} {units}"
        else:
            units_format_string = number_format_string
        formatter = cls(
            # number -> string uses given precision
            to_string=number_format_string.format,
            to_string_with_units=units_format_string.format,
            # number -> float just returns the number
            to_float=return_if_number,
        )
        return formatter

    @classmethod
    def for_ndarray(
        cls, precision: Optional[int], units: Optional[str]
    ) -> "ChannelFormatter":
        number_format_string = make_number_format_string(precision)

        # ndarray -> base64 encoded array
        def ndarray_to_base64_array(value: np.ndarray, length: int = 0) -> Base64Array:
            if length > 0:
                value = value[:length]
            return Base64Array(
                value.dtype.name.upper(), base64.b64encode(value.tobytes()).decode()
            )

        # ndarray -> [str] uses given precision
        def ndarray_to_string_array(value: np.ndarray, length: int = 0) -> List[str]:
            if length > 0:
                value = value[:length]
            func = number_format_string.format
            return [func(x) for x in value]

        formatter = cls(
            to_base64_array=ndarray_to_base64_array,
            to_string_array=ndarray_to_string_array,
        )

        return formatter

    @classmethod
    def for_enum(cls, choices: List[str]) -> "ChannelFormatter":
        formatter = cls(
            # enum string value
            to_string=choices.__getitem__,
            # enum index as a float
            to_float=float,
        )
        return formatter

    def __init__(
        self,
        to_string: Callable[[Any], str] = str,
        to_string_with_units: Callable[[Any], str] = str,
        to_float: Callable[[Any], Optional[float]] = return_none,
        to_base64_array: Callable[[Any, int], Optional[Base64Array]] = return_none,
        to_string_array: Callable[[Any, int], Optional[List[str]]] = return_none,
    ):
        self.to_string = to_string
        self.to_string_with_units = to_string_with_units
        self.to_float = to_float
        self.to_base64_array = to_base64_array
        self.to_string_array = to_string_array


@dataclass
class ChannelStatus:
    """
    The current status of a Channel, including alarm and connection status
    """

    # Of what quality is the current Channel value
    quality: str
    # Free form text describing the current status
    message: str
    # Whether the Channel will currently accept mutations
    mutable: bool

    @classmethod
    def valid(cls, mutable: bool = False) -> "ChannelStatus":
        return cls("VALID", "", mutable)

    @classmethod
    def warning(cls, message: str, mutable: bool = False) -> "ChannelStatus":
        return cls("WARNING", message, mutable)

    @classmethod
    def alarm(cls, message: str, mutable: bool = False) -> "ChannelStatus":
        return cls("ALARM", message, mutable)

    @classmethod
    def invalid(cls, message: str, mutable: bool = False) -> "ChannelStatus":
        return cls("INVALID", message, mutable)

    @classmethod
    def undefined(cls, message: str, mutable: bool = False) -> "ChannelStatus":
        return cls("UNDEFINED", message, mutable)

    @classmethod
    def changing(cls, message: str, mutable: bool = False) -> "ChannelStatus":
        return cls("CHANGING", message, mutable)


@strawberry.type
@dataclass
class ChannelTime:
    """
    Timestamp indicating when a value was last updated
    """

    # Floating point number of seconds since Jan 1, 1970 00:00:00 UTC
    seconds: float
    # A more accurate version of the nanoseconds part of the seconds field
    nanoseconds: int
    # An integer value whose interpretation is deliberately undefined
    userTag: int

    @classmethod
    def now(cls) -> "ChannelTime":
        now = time.time()
        return cls(now, int(now % 1 / 1e-9), 0)


@dataclass
class ChannelValue:
    value: Any
    formatter: ChannelFormatter = ChannelFormatter()


class Channel:
    def get_id(self) -> Optional[str]:
        raise NotImplementedError(self)

    def get_value(self) -> Optional[ChannelValue]:
        raise NotImplementedError(self)

    def get_time(self) -> Optional[ChannelTime]:
        raise NotImplementedError(self)

    def get_status(self) -> Optional[ChannelStatus]:
        raise NotImplementedError(self)

    def get_display(self) -> Optional[ChannelDisplay]:
        raise NotImplementedError(self)



import asyncio
import math
import time
from dataclasses import dataclass, replace
from typing import Any, AsyncGenerator, Dict, List, Optional, Sequence, Set, Type

import numpy as np

from coniql.coniql_schema import DisplayForm, Widget
from coniql.plugin import Plugin, PutValue
from coniql.types import (
    Channel,
    ChannelDisplay,
    ChannelFormatter,
    ChannelRole,
    ChannelStatus,
    ChannelTime,
    ChannelValue,
    Range,
)

# How long to keep Sim alive after the last listener has gone
SIM_DESTROY_TIMEOUT = 10

# Map of pv func to its Sim class
CHANNEL_CLASSES: Dict[str, Type["Sim"]] = {}


def register_channel(func: str):
    def decorator(cls: Type[Sim]):
        CHANNEL_CLASSES[func] = cls
        return cls

    return decorator


@dataclass
class SimChannel(Channel):
    id: Optional[str] = None
    value: Optional[ChannelValue] = None
    display: Optional[ChannelDisplay] = None
    time: Optional[ChannelTime] = None
    status: Optional[ChannelStatus] = None

    def get_id(self) -> Optional[str]:
        return self.id

    def get_value(self) -> Optional[ChannelValue]:
        return self.value

    def get_display(self) -> Optional[ChannelDisplay]:
        return self.display

    def get_time(self) -> Optional[ChannelTime]:
        return self.time

    def get_status(self) -> Optional[ChannelStatus]:
        return self.status


class Sim:
    def __init__(self, update_seconds: float):
        self.update_seconds = update_seconds
        self.channel = SimChannel(time=ChannelTime.now(), status=ChannelStatus.valid())

    def apply_changes(self, value, **changes) -> Channel:
        # pop changes that haven't really changed
        changes = {k: v for k, v in changes.items() if getattr(self.channel, k) != v}
        # value needs special treatment as might wrap a numpy array
        assert self.channel.value is not None, self.channel
        # Don't check numpy arrays as it's expensive, assume it changed
        if isinstance(value, np.ndarray) or value != self.channel.value.value:
            changes["value"] = ChannelValue(value, self.channel.value.formatter)
        # time always changes
        changes["time"] = ChannelTime.now()
        # replace our stored channel with an updated one
        self.channel = replace(self.channel, **changes)
        # a channel with our differences
        channel = SimChannel(**changes)
        return channel

    def compute_changes(self) -> Channel:
        raise NotImplementedError(self)


def make_display(
    min_value: float,
    max_value: float,
    warning_percent: float,
    alarm_percent: float,
    description: str,
    role: ChannelRole,
    widget: Widget,
) -> ChannelDisplay:
    assert max_value > min_value, "max_value %s is not > min_value %s" % (
        max_value,
        min_value,
    )
    display_range = max_value - min_value
    alarm_range = display_range * alarm_percent / 100
    warning_range = display_range * warning_percent / 100
    display = ChannelDisplay(
        description=description,
        role=role,
        widget=widget,
        controlRange=Range(min_value, max_value),
        displayRange=Range(min_value, max_value),
        alarmRange=Range(
            min_value + (display_range - alarm_range) / 2,
            max_value - (display_range - alarm_range) / 2,
        ),
        warningRange=Range(
            min_value + (display_range - warning_range) / 2,
            max_value - (display_range - warning_range) / 2,
        ),
        units="au",
        precision=5,
        form=DisplayForm.DEFAULT,
    )
    return display


@register_channel("sine")
class SineSim(Sim):
    """Create a simulated float sine value

    Args:
        min_value: The minimum output value
        max_value: The maximum output value
        steps: The number of steps taken to produce a complete sine wave
        update_seconds: The time between each step
        warning_percent: Percentage of the full range, outside this is warning
        alarm_percent: Percentage of the full range, outside this is alarm
    """

    def __init__(
        self,
        min_value: float = -5.0,
        max_value: float = 5.0,
        steps: float = 10.0,
        update_seconds: float = 1.0,
        warning_percent: float = 80.0,
        alarm_percent: float = 90.0,
    ):
        super().__init__(update_seconds)
        self.min = min_value
        self.range = max_value - min_value
        self.step = 2 * math.pi / max(steps, 1)
        self.x = 0.0
        display = make_display(
            min_value,
            max_value,
            warning_percent,
            alarm_percent,
            description="A Sine value generator",
            role=ChannelRole.RO,
            widget=Widget.TEXTUPDATE,
        )
        self.channel.display = display
        self.channel.value = ChannelValue(
            0,
            ChannelFormatter.for_number(display.precision, display.units),
        )

    def compute_changes(self):
        self.x += self.step
        value = self.min + (math.sin(self.x) + 1.0) / 2.0 * self.range
        assert self.channel.display is not None
        assert self.channel.display.alarmRange is not None
        assert self.channel.display.warningRange is not None
        if not self.channel.display.alarmRange.contains(value):
            status = ChannelStatus.alarm("Outside alarm range")
        elif not self.channel.display.warningRange.contains(value):
            status = ChannelStatus.warning("Outside warning range")
        else:
            status = ChannelStatus.valid()
        return self.apply_changes(value, status=status)


@register_channel("sinewave")
class SineWaveSim(Sim):
    """Create a simulated float waveform of a sine wave

    Args:
        period_seconds: The time between repetitions on the sinewave in time
        sample_wavelength: The wavelength of the output sinewave
        size: The size of the output waveform (min 10 elements)
        update_seconds: The time between each step
        min_value: The minimum output value
        max_value: The maximum output value
    """

    def __init__(
        self,
        period_seconds: float = 1.0,
        sample_wavelength: float = 10.0,
        size: float = 50.0,
        update_seconds: float = 1.0,
        min_value: float = -5.0,
        max_value: float = 5.0,
    ):
        super().__init__(update_seconds)
        self.min = min_value
        self.range = max_value - min_value
        self.period = max(period_seconds, 0.001)
        self.size = int(size)
        self.wavelength = sample_wavelength
        self.start = time.time()
        display = make_display(
            min_value,
            max_value,
            warning_percent=100.0,
            alarm_percent=100.0,
            description="A Sine waveform generator",
            role=ChannelRole.RO,
            widget=Widget.PLOTY,
        )
        self.channel.display = display
        self.channel.value = ChannelValue(
            np.zeros(self.size, dtype=np.float64),
            ChannelFormatter.for_ndarray(display.precision, display.units),
        )

    def compute_changes(self) -> Channel:
        t = time.time() - self.start
        x0 = t / self.period
        x = 2 * math.pi * (x0 + np.arange(self.size) / self.wavelength)
        value = self.min + (np.sin(x) + 1.0) / 2.0 * self.range
        return self.apply_changes(value)


@register_channel("rampwave")
class RampWaveSim(Sim):
    """Create a simulated float waveform of a ramping variable

    Args:
        size: The size of the output waveform
        update_seconds: The time between new waveform updates
        min_value: The minimum output value (inclusive)
        max_value: The maximum output value (exclusive)
        step: The step between each value in the ramp
    """

    def __init__(
        self,
        size: float = 10.0,
        update_seconds: float = 1.0,
        min_value: float = 0.0,
        max_value: float = 100.0,
        step: float = 1.0,
    ):
        super().__init__(update_seconds)
        self.size = int(size)
        # A single ramp waveform
        ramp = np.arange(min_value, max_value, step)
        self.ramp_length = len(ramp)
        # How many of these ramps in a single output waveform
        iterations = math.ceil(self.size / self.ramp_length)
        # Output waveform that we will take sliding slices out of
        self.ramps = np.tile(ramp, iterations + 1)
        self.i = 0

        display = make_display(
            min_value,
            max_value,
            warning_percent=100.0,
            alarm_percent=100.0,
            description="A ramp waveform generator",
            role=ChannelRole.RO,
            widget=Widget.PLOTY,
        )
        self.channel.display = display
        self.channel.value = ChannelValue(
            self.ramps[: self.size],
            ChannelFormatter.for_ndarray(display.precision, display.units),
        )

    def compute_changes(self) -> Channel:
        self.i += 1
        if self.i >= self.ramp_length:
            self.i = 0
        value = self.ramps[self.i : self.i + self.size]
        return self.apply_changes(value)


class SimPlugin(Plugin):
    def __init__(self) -> None:
        # {pv: Sim}
        self.sims: Dict[str, Sim] = {}
        # {pv: {queue_for_each_listener}}
        self.listeners: Dict[str, Set[asyncio.Queue[Channel]]] = {}
        # Set of asyncio tasks running
        self.task_references: Set[asyncio.Task[Any]] = set()

    async def _start_computing(self, pv: str):
        sim = self.sims[pv]
        next_compute = time.time()
        last_had_listeners = next_compute
        while next_compute - last_had_listeners < SIM_DESTROY_TIMEOUT:
            next_compute += sim.update_seconds
            await asyncio.sleep(next_compute - time.time())
            changes = sim.compute_changes()
            for q in self.listeners[pv]:
                last_had_listeners = next_compute
                await q.put(changes)
        # no-one listening, remove sim
        del self.sims[pv]
        del self.listeners[pv]

    async def get_channel(self, pv: str, timeout: float) -> Channel:
        if pv not in self.sims:
            if "(" in pv:
                assert pv.endswith(")"), "Missing closing bracket in %r" % pv
                func, param_str = pv[:-1].split("(", 1)
                parameters = [float(param.strip()) for param in param_str.split(",")]
            else:
                func = pv
                parameters = []
            cls = CHANNEL_CLASSES[func]
            inst = cls(*parameters)
            display = inst.channel.display
            assert display
            self.sims[pv] = inst
            self.listeners[pv] = set()
            task = asyncio.create_task(self._start_computing(pv))
            self.task_references.add(task)

            def _on_completion(t):
                self.task_references.remove(t)

            task.add_done_callback(_on_completion)

        return self.sims[pv].channel

    async def subscribe_channel(self, pv: str) -> AsyncGenerator[Channel, None]:
        q: asyncio.Queue[Channel] = asyncio.Queue()
        try:
            channel = await self.get_channel(pv, 0)
            self.listeners[pv].add(q)
            yield channel
            while True:
                yield await q.get()
        finally:
            self.listeners[pv].remove(q)

    async def put_channels(
        self, pvs: List[str], values: Sequence[PutValue], timeout: float
    ):
        raise RuntimeError(f"Cannot put {values!r} to {pvs}, as they aren't writeable")



import asyncio
from subprocess import Popen
from typing import Any, AsyncIterator, Dict, List, Optional, cast

import pytest
from aioca import Subscription
from strawberry import Schema

from coniql.app import create_schema
from coniql.caplugin import CAPlugin
from coniql.strawberry_schema import store_global

from .conftest import (
    PV_PREFIX,
    SUBSCRIPTION_TIMEOUT,
    base64_put_query,
    base64_put_query_result,
    check_put_timestamp,
    enum_get_query,
    enum_get_query_result,
    get_longout_subscription_query,
    get_ticking_subscription_result,
    ioc_creator,
    list_put_query,
    list_put_query_result,
    long_and_enum_put_query,
    long_and_enum_put_query_result,
    longout_get_query,
    longout_get_query_result,
    longout_str_get_query,
    longout_str_get_query_result,
    longout_subscription_result,
    nan_get_query,
    nan_get_query_result,
    ticking_subscription_query,
)


@pytest.fixture(scope="session")
def schema():
    return create_schema(False)


async def query_schema(schema: Schema, query: str):
    result = await schema.execute(query)
    return result.data


@pytest.mark.parametrize(
    "query, expected_result",
    [
        (longout_get_query, longout_get_query_result),
        (longout_str_get_query, longout_str_get_query_result),
        (enum_get_query, enum_get_query_result),
        (nan_get_query, nan_get_query_result),
    ],
    ids=["int_query", "str_query", "enum_query", "nan_query"],
)
async def test_schema_get_pv(
    ioc: Popen, schema: Schema, query: str, expected_result: str
):
    result = await query_schema(schema, query)
    assert result == expected_result


@pytest.mark.parametrize(
    "query, expected_result",
    [
        (long_and_enum_put_query, long_and_enum_put_query_result),
        (list_put_query, list_put_query_result),
        (base64_put_query, base64_put_query_result),
    ],
    ids=["long_and_enum_put", "list_put", "base64_put"],
)
async def test_schema_put_pv(
    ioc: Popen, schema: Schema, query: str, expected_result: str
):
    result = await query_schema(schema, query)
    assert result == expected_result
    check_put_timestamp(result)


@pytest.mark.asyncio
async def test_subscribe_disconnect(schema: Schema):
    pv_prefix = PV_PREFIX + "EXTRA:"
    ioc_process = ioc_creator(pv_prefix)
    query = get_longout_subscription_query(pv_prefix)
    results: List[Dict[str, Any]] = []
    resp = await schema.subscribe(query)
    assert isinstance(resp, AsyncIterator)
    async for result in resp:
        assert result.errors is None
        assert result.data
        if not results:
            # First response; now disconnect.
            results.append(result.data)
            ioc_process.communicate("exit()")
        else:
            # Second response; done.
            results.append(result.data)

            break
    assert len(results) == 2
    assert results[0] == longout_subscription_result[0]
    assert results[1] == longout_subscription_result[1]


@pytest.mark.asyncio
async def test_subscribe_ticking(ioc: Popen, schema: Schema):
    results: List[Optional[Dict[str, Any]]] = []
    resp = await schema.subscribe(ticking_subscription_query)
    assert isinstance(resp, AsyncIterator)
    count = 0
    while True:
        if count > 2:
            break
        # Set a timeout on wait for a response as otherwise this call will
        # block forever unless the schema.subscribe() receives data
        result = await asyncio.wait_for(resp.__anext__(), timeout=SUBSCRIPTION_TIMEOUT)
        results.append(result.data)
        count += 1
    assert len(results) == 3
    # Determine the starting value in the subscription
    startSub = results[0]
    assert startSub
    startVal = startSub["subscribeChannel"]["value"]["float"]
    subscription_result = get_ticking_subscription_result(startVal)
    for i in range(3):
        assert results[i] == subscription_result[i]


@pytest.mark.asyncio
async def test_subscribe_multiple(schema: Schema):
    """Test that multiple subscriptions to the same PV all receive the right data"""

    pv_prefix = PV_PREFIX + "MULTIPLE:"
    ioc_process = ioc_creator(pv_prefix)

    query = get_longout_subscription_query(pv_prefix)

    responses = await asyncio.gather(
        schema.subscribe(query), schema.subscribe(query), schema.subscribe(query)
    )

    # Check initial response is correct for all
    for resp in responses:
        results: List[Dict[str, Any]] = []
        assert isinstance(resp, AsyncIterator)
        async for result in resp:
            assert result.errors is None
            assert result.data
            results.append(result.data)
            break
        assert len(results) == 1
        assert results[0] == longout_subscription_result[0]

    ioc_process.communicate("exit()")

    # And check they all get the disconnect message
    for resp in responses:
        results = []
        assert isinstance(resp, AsyncIterator)
        async for result in resp:
            assert result.errors is None
            assert result.data
            results.append(result.data)
            break
        assert len(results) == 1
        assert results[0] == longout_subscription_result[1]


async def subscribe_task_wrapper(schema: Schema, query: str):
    resp = await schema.subscribe(query)
    assert isinstance(resp, AsyncIterator)
    async for _ in resp:
        pass


@pytest.mark.parametrize("num_subscribers", [1, 5])
async def test_subscribe_unsubscribe(ioc: Popen, schema: Schema, num_subscribers: int):
    """Test that cancelling a subscription correctly closes channel monitors"""

    # Create subscription(s)
    query = get_longout_subscription_query(PV_PREFIX)
    tasks: List[asyncio.Task] = []
    for _ in range(num_subscribers):
        tasks.append(asyncio.create_task(subscribe_task_wrapper(schema, query)))

    await asyncio.sleep(1)

    # Check the subscription is logged in the manager
    ca_plugin: CAPlugin = cast(CAPlugin, store_global.plugins["ca"])
    pvs = ca_plugin.subscription_manager.pvs
    assert len(pvs.keys()) == 1
    for pv in pvs.values():
        assert pv.meta_monitor.state == Subscription.OPEN
        assert pv.time_monitor.state == Subscription.OPEN
        assert pv.subscribers == num_subscribers

    # Close subscriptions
    for task in tasks:
        task.cancel()

    await asyncio.sleep(1)

    # Check subscription has been closed
    pvs = ca_plugin.subscription_manager.pvs
    for pv in pvs.values():
        assert pv.meta_monitor.state == Subscription.CLOSED
        assert pv.time_monitor.state == Subscription.CLOSED
        assert pv.subscribers == 0



import time
from typing import Dict, List, Optional

from aioca import Subscription, get_channel_infos
from aiohttp import web
from aiohttp.hdrs import ACCEPT
from aiohttp.web_request import Request
from aioprometheus import (
    REGISTRY,
    Counter,
    Gauge,
    Summary,
    count_exceptions,
    inprogress,
)
from aioprometheus.asgi.middleware import EXCLUDE_PATHS
from aioprometheus.renderer import render
from graphql import GraphQLError
from strawberry.aiohttp.handlers import GraphQLTransportWSHandler, GraphQLWSHandler
from strawberry.extensions import SchemaExtension
from strawberry.schema import Schema
from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL
from strawberry.types import ExecutionContext

# Create all the metrics for the entire application here
SUBSCRIPTIONS_IN_PROGRESS = Gauge(
    "coniql_subscriptions_in_progress", "Number of subscriptions in progress"
)
REQUEST_TIME = Summary(
    "coniql_request_processing_seconds", "Time spent processing request"
)
REQUESTS = Counter("coniql_request_total", "Total number of requests")
REQUEST_EXCEPTIONS = Counter(
    "coniql_request_handler_exceptions",
    "Number of exceptions and GraphQL errors in requests",
)
DROPPED_UPDATES = Gauge(
    "coniql_dropped_updates", "Number of updates dropped in subscriptions"
)
ACTIVE_CHANNELS = Gauge("coniql_active_channels", "Number of active channels in aioca")


class MetricsExtension(SchemaExtension):
    def on_operation(self):
        """Counts the number of GraphQL Queries and Mutations"""
        REQUESTS.inc({"route": "GraphQL"})
        yield


class MetricsSchema(Schema):
    """Extended Schema with metrics"""

    def process_errors(
        self,
        errors: List[GraphQLError],
        execution_context: Optional[ExecutionContext] = None,
    ):
        """Override to count the number of GraphQL errors"""
        labels = {"route": "GraphQL"}
        # Technically there could be multiple GraphQL errors present, but for metric
        # purposes we really only care that any happened.
        if len(errors):
            if (
                execution_context
                and execution_context.context
                and "request" in execution_context.context
            ):
                # If we can, add the remote end's address
                labels.update({"remote": execution_context.context["request"].remote})
            REQUEST_EXCEPTIONS.inc(labels)

        super().process_errors(errors, execution_context)


class MetricsGraphQLTransportWSHandler(GraphQLTransportWSHandler):
    """Custom override of GraphQLTransportWSHandler to allow adding the @inprogress
    annotation. Tracks how many subscriptions are currently active."""

    @inprogress(
        SUBSCRIPTIONS_IN_PROGRESS,
        labels={"type": f"subscription_{GRAPHQL_TRANSPORT_WS_PROTOCOL}"},
    )
    async def handle_request(self):
        await super().handle_request()


class MetricsGraphQLWSHandler(GraphQLWSHandler):
    """Custom override of GraphQLWSHandler to allow adding the @inprogress
    annotation. Tracks how many subscriptions are currently active."""

    @inprogress(
        SUBSCRIPTIONS_IN_PROGRESS,
        labels={"type": f"subscription_{GRAPHQL_WS_PROTOCOL}"},
    )
    async def handle_request(self):
        await super().handle_request()


def update_subscription_metrics(
    subscription: Subscription, last_dropped_count: int, labels: Dict[str, str]
) -> int:
    """Inspect the given Subscription to update the relevant metrics.
    Must provide the number of dropped updates prevously observed.
    Returns the current number of dropped updates."""
    labels.update({"pv": subscription.name})
    # Ensure only read dropped_callbacks once, in case it is updating in the background
    current_dropped_count = subscription.dropped_callbacks

    difference = current_dropped_count - last_dropped_count

    # Note that adding 0 will still cause the metric to be published
    DROPPED_UPDATES.add(labels, difference)
    return current_dropped_count


def update_active_channels() -> None:
    """Inspect aioca's channel cache to see how many active channels there are"""
    channel_infos = get_channel_infos()

    active_channels = sum(channel.connected for channel in channel_infos)

    ACTIVE_CHANNELS.set({}, active_channels)


async def handle_metrics(request: Request):
    """Create the output for all defined metrics, in a format determined by the
    request's ACCEPT headers."""
    update_active_channels()
    content, http_headers = render(REGISTRY, request.headers.getall(ACCEPT, []))
    return web.Response(body=content, headers=http_headers)


@web.middleware
@count_exceptions(
    REQUEST_EXCEPTIONS, {"route": "middleware"}
)  # Count any exceptions in any request
async def metrics_middleware(request: Request, handler):
    """Middleware that is called for all requests to the aiohttp server"""

    labels = {"route": "middleware", "path": request.path}

    # Ignore requests for some common paths.
    # Use the same list as used in aioprometheus for AGSI apps
    if request.path not in EXCLUDE_PATHS:
        REQUESTS.inc(labels)
        start_time = time.monotonic()

    response = await handler(request)

    if request.path not in EXCLUDE_PATHS:
        REQUEST_TIME.observe(labels, time.monotonic() - start_time)

    return response



import importlib
from unittest.mock import MagicMock

import pytest
from aiohttp.test_utils import TestClient
from aioprometheus import REGISTRY

import coniql.app
import coniql.metrics
from coniql.metrics import (
    ACTIVE_CHANNELS,
    DROPPED_UPDATES,
    REQUEST_EXCEPTIONS,
    REQUEST_TIME,
    REQUESTS,
    SUBSCRIPTIONS_IN_PROGRESS,
    update_subscription_metrics,
)

from .conftest import enum_get_query, longout_get_query, nan_get_query


@pytest.fixture(autouse=True)
def clear_metrics():
    """Reset all metrics before and after each metrics-based test"""
    REGISTRY.clear()
    # Must forcibly reload the modules in order to a) recreate the metrics and
    # b) re-create the "metrics_middleware" wrapped function that is used when
    # creating the application.
    importlib.reload(coniql.metrics)
    importlib.reload(coniql.app)
    yield
    REGISTRY.clear()


@pytest.mark.asyncio
async def test_metrics_endpoint(client: TestClient):
    """Test the metrics endpoint starts and returns basic info"""

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()
    expected_collectors = REGISTRY.collectors

    # Check the endpoint returns the same number of collectors as the registry
    assert text.count("HELP") == len(expected_collectors)
    assert text.count("TYPE") == len(expected_collectors)

    # And check that each named expected item exists
    assert ACTIVE_CHANNELS.name in text
    assert DROPPED_UPDATES.name in text
    assert REQUEST_EXCEPTIONS.name in text
    assert REQUEST_TIME.name in text
    assert REQUESTS.name in text
    assert SUBSCRIPTIONS_IN_PROGRESS.name in text


@pytest.mark.asyncio
async def test_metrics_query_count(ioc, client: TestClient):
    """Test metrics for counting GraphQL Queries and Mutations"""

    resp = await client.get("/ws", params={"query": longout_get_query})
    assert resp.status == 200

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()

    # The request total metrics
    assert 'coniql_request_total{path="/ws",route="middleware"} 1' in text
    assert 'coniql_request_total{route="GraphQL"} 1' in text


@pytest.mark.asyncio
async def test_metrics_processing_time(ioc, client: TestClient):
    """Test metrics for duration of requests.
    Note that this will only count the first request (/ws), as that is the only
    completed request at the time the metrics output are created. There will be no
    timing information for the /metrics call itself."""

    resp = await client.get("/ws", params={"query": longout_get_query})
    assert resp.status == 200

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()

    # Only expect 1 as the current request isn't tracked
    assert (
        "coniql_request_processing_seconds_count"
        '{path="/ws",route="middleware"} 1' in text
    )
    assert (
        "coniql_request_processing_seconds"
        '{path="/ws",quantile="0.5",route="middleware"}' in text
    )
    assert (
        "coniql_request_processing_seconds"
        '{path="/ws",quantile="0.9",route="middleware"}' in text
    )
    assert (
        "coniql_request_processing_seconds"
        '{path="/ws",quantile="0.99",route="middleware"}' in text
    )
    assert (
        "coniql_request_processing_seconds_sum"
        '{path="/ws",route="middleware"}' in text
    )


@pytest.mark.asyncio
async def test_metrics_active_channels(ioc, client: TestClient):
    """Test metrics for counting active channels in aioca"""

    # Open three Channels (three different PVs)
    resp = await client.get("/ws", params={"query": longout_get_query})
    assert resp.status == 200
    resp = await client.get("/ws", params={"query": enum_get_query})
    assert resp.status == 200
    resp = await client.get("/ws", params={"query": nan_get_query})
    assert resp.status == 200

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()

    assert "coniql_active_channels 3" in text


@pytest.mark.asyncio
async def test_metrics_graphql_exception(client: TestClient):
    """Test metrics for counting GraphQL errors"""

    # Deliberately send a malformed query
    resp = await client.get("/ws", params={"query": longout_get_query[0:100]})
    assert resp.status == 200

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()

    assert (
        'coniql_request_handler_exceptions{remote="127.0.0.1",route="GraphQL"} 1'
        in text
    )


@pytest.mark.asyncio
async def test_metrics_aiohttp_exception(client: TestClient):
    """Test metrics for counting aiohttp errors"""

    resp = await client.get("/fake_endpoint")
    assert resp.status == 404

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()

    assert 'coniql_request_handler_exceptions{route="middleware"} 1' in text


@pytest.mark.asyncio
async def test_metrics_mocked_dropped_updates(client: TestClient):
    """Test metrics for dropped subscriptions by using mocked methods.
    This is done because it's very difficult to deliberately make aioca miss updates"""

    mocked_subscription = MagicMock()
    mocked_subscription.name = "TEST-NAME"
    mocked_subscription.dropped_callbacks = 5

    last_dropped = 0
    last_dropped = update_subscription_metrics(mocked_subscription, last_dropped, {})
    assert last_dropped == 5

    resp = await client.get("/metrics")
    assert resp.status == 200

    text = await resp.text()

    assert 'coniql_dropped_updates{pv="TEST-NAME"} 5' in text

    last_dropped = update_subscription_metrics(mocked_subscription, last_dropped, {})
    assert last_dropped == 5


@pytest.mark.asyncio
async def test_metrics_subscriptions_in_progress(
    ioc, client: TestClient, subscription_data
):
    """Test metrics for subscriptions in progress"""
    ws_protocol, msg_init, msg_ack, msg_send = subscription_data

    async with client.ws_connect("/ws", protocols=[ws_protocol]) as ws:
        await ws.send_json(msg_init)
        response = await ws.receive_json()
        assert response == msg_ack

        await ws.send_json(msg_send)

        resp = await client.get("/metrics")
        assert resp.status == 200

        text = await resp.text()
        assert (
            f'coniql_subscriptions_in_progress{{type="subscription_{ws_protocol}"}} 1'
            in text
        )

        await ws.close()
        assert ws.closed


@pytest.mark.asyncio
async def test_metrics_openmetrics_format(client: TestClient):
    """Test that the metrics endpoint can return data in openmetrics format, which
    is what Prometheus expects"""
    resp = await client.get(
        "/metrics", headers={"Accept": "application/openmetrics-text;version=1.0.0"}
    )
    assert resp.status == 200

    text = await resp.text()

    # Check an arbitrary line is in the right format
    assert "# TYPE coniql_active_channels gauge" in text



from typing import AsyncIterator, Dict, List, Sequence, Tuple, Union

import numpy as np

from coniql.types import Channel

PutValue = Union[bool, int, float, str, List[str], np.ndarray]


class Plugin:
    transport: str

    async def get_channel(self, pv: str, timeout: float) -> Channel:
        """Get the current structure of a Channel"""
        raise NotImplementedError(self)

    async def put_channels(
        self, pvs: List[str], values: Sequence[PutValue], timeout: float
    ):
        """Put a value to a channel, returning the structure after put"""
        raise NotImplementedError(self)

    async def subscribe_channel(self, pv: str) -> AsyncIterator[Channel]:
        """Subscribe to the structure of the Channel, yielding structures
        where only changing top level fields are filled in"""
        raise NotImplementedError(self)
        yield


class PluginStore:
    def __init__(self) -> None:
        # {transport: plugin}
        self.plugins: Dict[str, Plugin] = {}

    def add_plugin(self, transport: str, plugin: Plugin, set_default=False):
        self.plugins[transport] = plugin
        plugin.transport = transport
        if set_default:
            self.plugins[""] = plugin

    def transport_pv(self, channel_id: str) -> Tuple[str, str]:
        """Take a channel_id with an optional transport prefix and
        return the transport and pv components"""
        split = channel_id.split("://", 1)
        if len(split) == 1:
            transport, pv = self.plugins[""].transport, channel_id
        else:
            transport, pv = split
        return transport, pv

    def plugin_config_id(self, channel_id: str) -> Tuple[Plugin, str]:
        transport, pv = self.transport_pv(channel_id)
        channel_id = f"{transport}://{pv}"
        plugin = self.plugins[transport]
        return plugin, channel_id



import argparse
import asyncio
import datetime
import sys
import threading
import time
import uuid
from typing import Dict

import ujson
import websockets

# Constants
cpu_average = 0
memory_use = 0
PV_PREFIX = "TEST:REC"
subscriptions_list: Dict[str, "PVSubscription"] = {}
task_list = []


parser = argparse.ArgumentParser(description="Process inputs")
parser.add_argument(
    "-n", "--npvs", action="store", dest="n_pvs", default=1, help="Number of PVs"
)
parser.add_argument(
    "-s",
    "--nsamples",
    action="store",
    dest="n_samples",
    default=10,
    help="Number of samples to collect",
)
parser.add_argument(
    "-p",
    "--protocol",
    action="store",
    dest="ws_protocol",
    choices=["1", "2"],
    default=1,
    help="websocket protocol: 1 = graphql-ws, 2 = graphql-transport-ws",
)
parser.add_argument(
    "-f",
    "--file",
    action="store",
    dest="output_file",
    default="performance_test_results.txt",
    help="File to output results to",
)
parser.add_argument(
    "-l",
    "--log-file",
    action="store",
    dest="log_filename",
    default=None,
    help="File to log subscription progress to",
)
parser.add_argument(
    "-x",
    "--no-cpu-monitor",
    action="store_true",
    dest="no_cpu_monitor",
    help="Do not run the CPU monitor",
)
parser.add_argument(
    "--coniql-addr",
    action="store",
    dest="addr",
    help="Address of Coniql server websocket",
    default="0.0.0.0:8080/ws",
)
parser.add_argument(
    "--debug", action="store_true", help="Enable extra debugging information"
)


class GraphQLClient:
    def __init__(self, endpoint, signal, ws_protocol, log_filename):
        self.endpoint = endpoint
        self.signal = signal
        self.ws_protocol = ws_protocol
        self.first_subscribe = True
        self.log_filename = log_filename

        if self.log_filename is not None:
            print("-> Logging subscription progress")
            self.log_file = open(self.log_filename, "w")
        else:
            self.log_file = None

    async def subscribe(self, idid, query, handle, n_messages, debug):
        monitor_progress = False
        if self.first_subscribe and self.log_filename is not None:
            monitor_progress = True
            self.first_subscribe = False

        if debug:
            monitor_progress = True

        uid = str(uuid.uuid4())
        connection_init_message = ujson.dumps(
            {"type": "connection_init", "payload": {}}
        )

        request_message_graphql_ws_protocol = ujson.dumps(
            {"type": "start", "id": uid, "payload": {"query": query}}
        )

        request_message_graphql_transport_ws_protocol = ujson.dumps(
            {"type": "subscribe", "id": uid, "payload": {"query": query}}
        )

        stop_message_graphql_ws_protocol = ujson.dumps({"type": "stop", "id": uid})

        stop_message_graphql_transport_ws_protocol = ujson.dumps(
            {"type": "complete", "id": uid}
        )

        protocols = ["graphql-ws", "graphql-transport-ws"]

        async with websockets.connect(
            self.endpoint,
            subprotocols=[protocols[self.ws_protocol - 1]],
        ) as websocket:
            await websocket.send(connection_init_message)
            if self.ws_protocol == 2:
                await websocket.send(request_message_graphql_transport_ws_protocol)
            else:
                await websocket.send(request_message_graphql_ws_protocol)

            msg_count = 0
            start_time = time.time()
            async for response in websocket:
                data = ujson.loads(response)
                if data["type"] == "connection_ack":
                    pass
                elif data["type"] == "ka":
                    pass
                else:
                    if self.signal.get_start():
                        handle(data["payload"])
                        if n_messages is None:
                            # Do nothing and continue subscription indefinitely
                            pass
                        else:
                            if msg_count > n_messages:
                                if self.ws_protocol == 2:
                                    await websocket.send(
                                        stop_message_graphql_ws_protocol
                                    )
                                else:
                                    await websocket.send(
                                        stop_message_graphql_transport_ws_protocol
                                    )

                                if self.log_file and not self.log_file.closed:
                                    self.log_file.close()
                                break
                            msg_count = msg_count + 1
                            if monitor_progress:
                                ten_percent = round(n_messages / 10, 0)
                                if msg_count % ten_percent == 0:
                                    remaining_time = (
                                        (time.time() - start_time)
                                        / msg_count
                                        * (n_messages - msg_count)
                                    )

                                    message = (
                                        f"Collected {msg_count}/{n_messages} samples "
                                        f"for pv {idid}. "
                                        f"Remaining time: {round(remaining_time, 0)}"
                                        "secs \n"
                                    )

                                    print(message)
                                    # The file may be closed by another instance of
                                    # this function.
                                    if self.log_file and not self.log_file.closed:
                                        self.log_file.write(message)
                                        self.log_file.flush()
                    else:
                        continue


class StartStopSignal:
    def __init__(self):
        self.start = False
        self.stop = False

    def signal_start(self):
        print("-> Starting monitor")
        self.start = True

    def signal_stop(self):
        print("-> Stopping monitor")
        self.stop = True

    def get_start(self):
        return self.start

    def get_stop(self):
        return self.stop


class PVSubscription:
    def __init__(self, pv):
        self.pv = pv
        self.values = []

    def append(self, value):
        self.values.append(value)

    def get_values(self):
        return self.values


def cpu_monitor(signal):
    # Keep import local so we don't need it in Container
    import psutil

    pid = 0
    for proc in psutil.process_iter(["pid", "name"]):
        if proc.info["name"] == "coniql":
            pid = proc.info["pid"]
            print("-> Monitoring PID: " + str(pid))

    p = psutil.Process(pid)
    mem = p.memory_info().rss / 1000000
    memi = mem
    cpu_res = []
    count = 0
    while True:
        if signal.get_stop():
            break
        if signal.get_start():
            cpu = p.cpu_percent(interval=1.0)
            mem = p.memory_info().rss / 1000000
            print("-> CPU: " + str(cpu) + ", MEM: " + str(mem))
            cpu_res.append(cpu)
            if count == 0:
                memi = mem
            count += 1
        time.sleep(0.1)

    memf = mem
    mem_use = memf - memi
    # Remove last element which may have been taken after subscriptions finished
    cpu_res.pop()
    if len(cpu_res) > 0:
        cpu_aver = sum(cpu_res) / (len(cpu_res))
        global cpu_average
        cpu_average = cpu_aver
        global memory_use
        memory_use = mem_use


def data_handler(data):
    id = data["data"]["subscribeChannel"]["id"]
    id = id.replace("ca://", "")

    # Add value to list
    value = data["data"]["subscribeChannel"]["value"]["float"]
    subscriptions_list[id].append(value)


def get_subscription_query(pv_name):
    return (
        """subscription {
  subscribeChannel(id: "ca://%s") {
    id
    time {
        datetime
    }
    value {
        string
        float
        base64Array {
          numberType
          base64
        }
      stringArray
      }
      status {
        quality
        message
        mutable
      }
      display {
        units
        form
        controlRange {
          max
          min
        }
        choices
        precision
      }
    }
  }"""
        % pv_name
    )


async def coniql_subscription(client: GraphQLClient, pv_name, n_samples, debug):
    subscriptions_list[pv_name] = PVSubscription(pv_name)
    await client.subscribe(
        idid=pv_name,
        query=get_subscription_query(pv_name),
        handle=data_handler,
        n_messages=n_samples,
        debug=debug,
    )


async def main():
    args = parser.parse_args()
    n_pvs = int(args.n_pvs)
    n_samples = int(args.n_samples)
    ws_protocol = int(args.ws_protocol)
    log_filename = args.log_filename
    no_cpu_monitor = args.no_cpu_monitor
    debug = args.debug

    protocol = "graphql-ws"
    if ws_protocol == 2:
        protocol = "graphql-transport-ws"
    print("-> Using the websocket protocol: '" + protocol + "'")

    # Create client
    signal = StartStopSignal()
    client = GraphQLClient(
        endpoint=f"ws://{args.addr}",
        signal=signal,
        ws_protocol=ws_protocol,
        log_filename=log_filename,
    )
    # Start CPU monitor thread
    if no_cpu_monitor:
        print("-> Not running CPU monitor")
    else:
        cpu_monitor_thread = threading.Thread(target=cpu_monitor, args=(signal,))
        cpu_monitor_thread.start()

    # Create subscription tasks for n_pvs
    for i in range(n_pvs):
        # Get the PV name
        pv_name = PV_PREFIX + str(i)

        # Create a task
        sub_task = asyncio.create_task(
            coniql_subscription(client, pv_name, n_samples, debug)
        )

        # Add to list of tasks to await
        task_list.append(sub_task)
        print("-> Starting subscription: " + str(pv_name))

    # Signal for the CPU monitor thread to start recording CPU metrics
    signal.signal_start()

    start_time = time.time()

    # Await all subscriptions to complete
    try:
        await asyncio.gather(*task_list)
        exit_code = 0
    except Exception as e:
        # Catch any exceptions so that we can still record results
        print("Exception caught: ", e)
        print("repr: ", repr(e))  # Sometimes just printing "e" is blank...
        exit_code = 1
    finally:
        # Signal to CPU monitoring thread to stop recording CPU metrics
        signal.signal_stop()

    duration = time.time() - start_time

    # Analyse results
    missing_average = 0
    missing_max = 0
    for pv in subscriptions_list.keys():
        res = subscriptions_list[pv].get_values()
        if len(res) == 0:
            break
        expected_result = res[0]

        if expected_result is None:
            print("ERROR: Found None for ", pv, " ", res)

        missing = 0
        for val in res:
            if val != expected_result:
                n_missed = val - expected_result
                missing = missing + n_missed
                expected_result = val + 1
            else:
                expected_result += 1

        if missing > missing_max:
            missing_max = missing
        missing_average = missing_average + missing / n_pvs
        sample_range = max(res) - min(res) + 1
        print(
            pv
            + " processing complete: value range "
            + str(sample_range)
            + ", missing "
            + str(missing)
        )

    # Collect results
    time.sleep(1)

    # NOTE: Kubernetes performance test relies on these being the last 3 lines
    # of output when not using CPU monitor
    print("\n\n ****** SUMMARY ******")
    print(" Average missed events = " + str(round(missing_average)))
    print(" Max. missed events = " + str(missing_max))

    if not no_cpu_monitor:
        res_str = (
            f"[{datetime.datetime.now()}](nPVs={n_pvs}, nsamples={n_samples}, "
            f"protocol={protocol})| Av. missed events: {round(missing_average)}| "
            f"Max missed events: {missing_max}| CPU av.: {cpu_average:.2f} %| "
            f"Mem usage: {memory_use:.2f} MiB| Duration: {duration:.2f} secs\n"
        )

        with open(args.output_file, "a") as f:
            f.write(res_str)

        print(" CPU average: " + str(cpu_average) + " %")
        print(" Memory usage: " + str(memory_use) + " MiB")
        print(" *********************\n")
    sys.exit(exit_code)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import time
from subprocess import Popen
from typing import Any, Dict, List, Optional

import pytest
from aiohttp.test_utils import TestClient
from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL
from strawberry.subscriptions.protocols.graphql_transport_ws.types import (
    ConnectionAckMessage,
    ConnectionInitMessage,
    SubscribeMessage,
    SubscribeMessagePayload,
)
from strawberry.subscriptions.protocols.graphql_ws import GQL_CONNECTION_KEEP_ALIVE

from .conftest import (
    PV_PREFIX,
    SUBSCRIPTION_TIMEOUT,
    base64_put_query,
    base64_put_query_result,
    check_put_timestamp,
    enum_get_query,
    enum_get_query_result,
    get_longout_subscription_query,
    get_ticking_subscription_result,
    ioc_cleanup,
    ioc_creator,
    list_put_query,
    list_put_query_result,
    long_and_enum_put_query,
    long_and_enum_put_query_result,
    longout_get_query,
    longout_get_query_result,
    longout_str_get_query,
    longout_str_get_query_result,
    longout_subscription_result,
    nan_get_query,
    nan_get_query_result,
)


@pytest.mark.parametrize(
    "query, expected_result",
    [
        (longout_get_query, longout_get_query_result),
        (longout_str_get_query, longout_str_get_query_result),
        (enum_get_query, enum_get_query_result),
        (nan_get_query, nan_get_query_result),
    ],
    ids=["int_query", "str_query", "enum_query", "nan_query"],
)
async def test_aiohttp_client_get_pv(
    ioc: Popen, client: TestClient, query: str, expected_result: str
):
    resp = await client.get("/ws", params={"query": query})
    assert resp.status == 200
    result = await resp.json()
    assert result["data"] == expected_result


@pytest.mark.parametrize(
    "query, expected_result",
    [
        (long_and_enum_put_query, long_and_enum_put_query_result),
        (list_put_query, list_put_query_result),
        (base64_put_query, base64_put_query_result),
    ],
    ids=["long_and_enum_query", "list_query", "base64_query"],
)
async def test_aiohttp_client_put_pv(
    ioc: Popen, client: TestClient, query: str, expected_result: str
):
    resp = await client.post("/ws", json={"query": query})
    result = await resp.json()
    assert result["data"] == expected_result
    check_put_timestamp(result["data"])


@pytest.mark.asyncio
async def test_subscribe_disconnect(client: TestClient):
    pv_prefix = PV_PREFIX + "EXTRA:"
    ioc_process = ioc_creator(pv_prefix)
    query = get_longout_subscription_query(pv_prefix)
    results: List[Dict[str, Any]] = []
    async with client.ws_connect(
        "/ws", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        await ws.send_json(ConnectionInitMessage().as_dict())

        response = await ws.receive_json()
        assert response == ConnectionAckMessage().as_dict()

        await asyncio.sleep(0.1)
        await ws.send_json(
            SubscribeMessage(
                id="sub1",
                payload=SubscribeMessagePayload(query=query),
            ).as_dict()
        )
        while True:
            result = await ws.receive_json()
            if not results:
                # First response; now disconnect.
                results.append(result["payload"]["data"])
                ioc_process.communicate("exit()")
            else:
                # Second response; done.
                results.append(result["payload"]["data"])
                break

        await ws.close()
        assert ws.closed

    assert len(results) == 2
    assert results[0] == longout_subscription_result[0]
    assert results[1] == longout_subscription_result[1]
    ioc_cleanup(ioc_process)


@pytest.mark.asyncio
async def test_subscribe_pv(ioc: Popen, client: TestClient, subscription_data):
    results: List[Optional[Dict[str, Any]]] = []

    ws_protocol, msg_init, msg_ack, msg_send = subscription_data

    async with client.ws_connect("/ws", protocols=[ws_protocol]) as ws:
        await ws.send_json(msg_init)
        response = await ws.receive_json()
        assert response == msg_ack
        await ws.send_json(msg_send)
        count = 0
        start = time.time()
        while True:
            if time.time() - start > SUBSCRIPTION_TIMEOUT:
                pytest.fail("Timeout waiting for subscription data")
            if count > 2:
                break

            # Set a timeout on wait for a websocket response as this blocks for the
            # GRAPHQL_TRANSPORT_WS_PROTOCOL. A "keep alive" response is constantly
            # sent for the GRAPHQL_WS_PROTOCOL so still need the above timeout to catch
            # this case.
            result = await asyncio.wait_for(
                ws.receive_json(), timeout=SUBSCRIPTION_TIMEOUT
            )
            if result["type"] == GQL_CONNECTION_KEEP_ALIVE:
                continue
            results.append(result["payload"]["data"])
            count += 1

        await ws.close()
        assert ws.closed
    assert len(results) == 3
    # Determine the starting value in the subscription
    startSub = results[0]
    assert startSub
    startVal = startSub["subscribeChannel"]["value"]["float"]
    subscription_result = get_ticking_subscription_result(startVal)
    for i in range(3):
        assert results[i] == subscription_result[i]
